#! /usr/bin/python3
# -*- coding: utf-8 -*-

#####################################################################
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#####################################################################

class DecodeError(Exception):
    """Error occured during decode process."""

    def __init__(self, val=None, msg=None):
        self.val = val
        self.msg = msg


class EncodeError(Exception):
    """Error occured during encoding process."""

    def __init__(self, val=None, msg=None):
        self.val = val
        self.msg = msg

class FilePathError(Exception):
    """Bad path error."""

    def __init__(self, obj=None, msg=None):
        self.obj = obj
        self.msg = msg



#! /usr/bin/python3
# -*- coding: utf-8 -*-

#####################################################################
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#####################################################################

import os
import re
import json
from pyben.exceptions import FilePathError, DecodeError, EncodeError

class Bendecoder:
    """Decode class contains all decode methods."""

    def __init__(self, data=None):
        """
        Initialize instance with optional pre compiled data.

        Args
        --------
            data (bytes-like, optional): target data for decoding.
        """
        self.data = data
        self.decoded = None
        self.json = None

    @classmethod
    def load(cls, item):
        """
        Bendecoder.load(item)

        Extract contents from path/path-like and return Decoded data.

        Args
        --------
        path : str or IOBuffer
            Path containing bencoded data.

        Raises
        ---------
        FilePathError
            Incorrect path or IOBuffer doesnt exist.

        Returns
        ---------
        any
            Decoded contents of file, Usually a dictionary.
        """
        decoder = cls()
        if hasattr(item, "read"):
            data = item.read()
            return decoder.decode(data)
        elif os.path.exists(item) and os.path.isfile(item):
            with open(item, "rb") as fd:
                data = fd.read()
                return decoder.decode(data)
        raise FilePathError()

    @classmethod
    def loads(cls, data):
        """
        Benencoder.loads(data)

        Shortcut to Decode raw bencoded data.

        Args
        --------
        data : bytes
            Bendencoded bytes.

        Returns
        ---------
        any
            Decoded data usually a dictionary.
        """
        decoder = cls()
        return decoder.decode(data)

    def decode(self, data=None):
        """
        Decode bencoded data.

        Args
        -------
        bits : bytes
            bencoded data for decoding.

        Returns
        --------
        any
            the decoded data.
        """
        data = self.data if not data else data
        self.decoded, _ = self._decode(bits=data)
        return self.decoded

    def json(self, bits=None):
        """Return output of the decoder json serialized."""
        if not bits:
            if self.decoded is None:
                self.decoded = self.decode(self.data)
        else:
            self.decoded = self.decode(bits)
        self.json = json.dumps(self.decoded)
        return self.json

    def _decode(self, bits=None):
        if bits is None:
            bits = self.data
        if bits.startswith(b"i"):
            match, feed = self._decode_int(bits)
            return match, feed

        # decode string
        elif chr(bits[0]).isdigit():
            num, feed = self._decode_str(bits)
            return num, feed

        # decode list and contents
        elif bits.startswith(b"l"):
            lst, feed = self._decode_list(bits)
            return lst, feed

        # decode dictionary and contents
        elif bits.startswith(b"d"):
            dic, feed = self._decode_dict(bits)
            return dic, feed
        else:
            raise DecodeError(bits, "Unrecognized data. Cannot decode.")

    def _decode_dict(self, bits):
        """
        Decode keys and values in dictionary.

        Args
        --------
        bits (bytearray): bytes of data for decoding.

        Returns
        ---------
        [dict]
            : dictionary and contents.
        """
        dic, feed = {}, 1
        while not bits[feed:].startswith(b"e"):
            match1, rest = self._decode(bits[feed:])
            feed += rest
            match2, rest = self._decode(bits[feed:])
            feed += rest
            dic[match1] = match2
        feed += 1
        return dic, feed

    def _decode_list(self, bits):
        """
        Decode list and its contents.

        Args
        --------
        bits : bytearray
            Bencoded data.

        Returns
        ---------
        list
            : decoded list and contents
        """
        lst, feed = [], 1
        while not bits[feed:].startswith(b"e"):
            match, rest = self._decode(bits[feed:])
            lst.append(match)
            feed += rest
        feed += 1
        return lst, feed

    def _decode_str(self, bits):
        """
        Decode string.

        Args
        --------
        bits (bytearray): bencoded string.

        Returns
        ---------
        [str]
            : decoded string.
        """
        match = re.match(br"(\d+):", bits)
        word_len, start = int(match.groups()[0]), match.span()[1]
        word = bits[start : start + word_len]
        try:
            word = word.decode("utf-8")
        except:
            word = word.hex()
        return word, start + word_len

    def _decode_int(self, bits):
        """
        Decode intiger.

        Args
        --------
        bits (bytearray): bencoded intiger.

        Returns
        ---------
        [int]
            : decoded intiger.
        """
        obj = re.match(br"i(-?\d+)e", bits)
        return int(obj.group(1)), obj.end()


class Benencoder:
    """Encode collection of methods for Bencoding data."""

    def __init__(self, data=None):
        """
        Initialize Benencoder insance with optional pre compiled data.

        Args
        --------
        data : any, optional
            Target data for encoding. Defaults to None.
        """
        self.data = data
        self.encoded = None

    @classmethod
    def dump(cls, data, path):
        """
        Benecoder.dump(data,path)

        Shortcut Classmethod for encoding data and writing to file. This

        Args
        --------
        data : any
            Raw data to be encoded, usually dict.txt
        path : path-like or iobuffer
            Where encoded data should be written to.py

        Returns
        ---------
        bool:
             Return True if success.txt
        """
        encoded = cls(data).encode()
        if hasattr(path, "write"):
            path.write(encoded)
        else:
            with open(path, "wb") as fd:
                fd.write(encoded)
        return True

    @classmethod
    def dumps(cls, data):
        """
        self.dumps(data)

        Shortcut method for encoding data and immediately returning it.

        Args
        --------
        data : any
            Raw data to be encoded usually a dictionary.

        Returns
        ---------
        bytes:
             encoded data.
        """
        return cls(data).encode()

    def encode(self, val=None):
        """
        self.encode(val)

        Encodes data provided as an arguement or provided at initialization.

        Args
        --------
        val : any, optional
            Data for encoding. Defaults to None.

        Returns
        ---------
        bytes:
             encoded data
        """
        if val is None:
            val = self.data
        self.encoded = self._encode(val)
        return self.encoded

    def _encode(self, val):
        """
        self._encode(val)

        Encode data with bencode protocol.

        args
        --------
        bits : bytes
            Bencoded data for decoding.

        returns
        ---------
        any:
             the decoded data.
        """
        if type(val) == str:
            return self._encode_str(val)

        if hasattr(val, "hex"):
            return self._encode_bytes(val)

        elif type(val) == int:
            return self._encode_int(val)

        elif type(val) == list:
            return self._encode_list(val)

        elif type(val) == dict:
            return self._encode_dict(val)

        raise EncodeError(val)

    def _encode_bytes(self, val):
        size = str(len(val)) + ":"
        return size.encode("utf-8") + val

    def _encode_str(self, txt):
        """
        _encode_str(self, txt)

        Decode string.

        Args
        --------
        txt (str): string.

        Returns
        ---------
        [bytes]
            : bencoded string.
        """
        size = str(len(txt)).encode("utf-8")
        return size + b":" + txt.encode("utf-8")

    def _encode_int(self, i):
        """
        _encode_int(self,i)

        Encode intiger.

        Args
        --------
        i : int
            Intiger

        Returns
        ---------
        bytes
            : bencoded intiger.
        """
        return b"i" + str(i).encode("utf-8") + b"e"

    def _encode_list(self, elems):
        """Encode list and its contents.

        Args
        --------
        elems : list
            List of content to be encoded.

        Returns
        ---------
        bytes
            : bencoded data
        """
        lst = [b"l"]
        for elem in elems:
            encoded = self._encode(elem)
            lst.append(encoded)
        lst.append(b"e")
        bit_lst = b"".join(lst)
        return bit_lst

    def _encode_dict(self, dic):
        """
        Encode keys and values in dictionary.

        Args
        --------
        dic : dict
            Dictionary of data for encoding.

        Returns
        ---------
        bytes
            Bencoded data.
        """
        result = b"d"
        for k, v in dic.items():
            result += b"".join([self._encode(k), self._encode(v)])
        return result + b"e"



#! /usr/bin/python3
# -*- coding: utf-8 -*-

#####################################################################
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#####################################################################

import re
from pyben.exceptions import DecodeError, EncodeError

"""
Library of functions and classes for encoding and decoding bencoded data.

Usage: How to Encode/Decode:

>>> import pyben
>>> fd = "path/to/file"
>>> data = {"item1": ["item2", 3, [4], {5: "item6"}]}
>>> encoded = pyben.dumps(data)
>>> encoded
b'd5:item1l5:item2i3eli4eedi5e5:item6eee'
>>> decoded = pybem.loads(encoded)
{'item1': ['item2', 3, [4], {5: 'item6'}]}

"""

def dump(obj, iobuffer):
    """
    dump(obj, iobuffer)

    Shortcut function for bencode encode data and write to file.

    Args
    -------
    * obj : any
        Data to be encoded.
    * iobuffer : IOBuffer
        File of path-like to write the data to.
    """
    encoded = benencode(obj)
    if hasattr(iobuffer, "write"):
        iobuffer.write(encoded)
        return iobuffer.close()
    else:
        with open(iobuffer, "wb") as fd:
            fd.write(encoded)
    return iobuffer


def dumps(obj):
    """
    dumps(obj)

    Shortuct function to encoding given obj to bencode encoding.

    Args
    --------
    obj : any
        object to be encoded.py

    Returns
    ---------
    bytes:
         encoded data
    """
    return benencode(obj)

def load(iobuffer):
    """
    load(iobuffer)

    Loads bencoded data from a file of path object and decodes it.

    Args
    --------
    iobuffer : path or file
        Open and/or read data from file to be decoded

    Returns
    ---------
    (any)
        usually a dictionary... decoded data
    """
    if hasattr(iobuffer, "read"):
        decoded, _ = bendecode(iobuffer.read())
    else:
        with open(iobuffer, "rb") as fd:
            decoded, _ = bendecode(fd.read())
    return decoded

def loads(encoded):
    """
    loads(encoded)

    Shortcut function for decoding encoded data.py.

    Args
    --------
    encoded : bytes
        Bencoded data.

    Returns
    ---------
    any:
         usually a dictionary, decoded data.
    """
    decoded, _ = bendecode(encoded)
    return decoded


def bendecode_str(bits):
    """
    bendecode_str(bits)

    Bendecode string types.

    Args
    --------
    bits : bytes
        Bencoded string

    Returns
    ---------
    str:
         Decoded data string
    """
    match = re.match(br"(\d+):", bits)
    word_len, start = int(match.groups()[0]), match.span()[1]
    word = bits[start : start + word_len]
    try:
        word = word.decode("utf-8")
    except:
        word = word.hex()
    return word, start + word_len


def bendecode_int(bits):
    """
    bendecode_int

    Decode digits.

    Args
    --------
    bits : bytes
        Bencoded intiger bytes

    Returns
    ---------
    int:
         decoded int value
    """
    obj = re.match(br"i(-?\d+)e", bits)
    return int(obj.group(1)), obj.end()


def bendecode_dict(bits):
    """
    bendecode_dict

    Decode dictionary and it's contents.

    Args
    --------
    bits : bytes
        Bencoded dictionary.

    Returns
    ---------
    dict
        Decoded dictionary and contents
    """
    dic, feed = {}, 1
    while not bits[feed:].startswith(b"e"):
        match1, rest = bendecode(bits[feed:])
        feed += rest
        match2, rest = bendecode(bits[feed:])
        feed += rest
        dic[match1] = match2
    feed += 1
    return dic, feed


def bendecode_list(bits):
    """
    bendecode_list

    Decode list and list contents.

    Args
    --------
    bits : bytes
        Bencoded list.

    Returns
    ---------
    list:
         decoded list and contents.
    """
    lst, feed = [], 1
    while not bits[feed:].startswith(b"e"):
        match, rest = bendecode(bits[feed:])
        lst.append(match)
        feed += rest
    feed += 1
    return lst, feed


def bendecode(bits):
    """
    bendecode

    Decode bencoded data.

    Args
    --------
    bits : bytes
        Bencoded data.

    Raises
    --------
    Exception
        Malformed data.

    Returns
    ---------
    any:
         decoded data.
    """
    if bits.startswith(b"i"):
        match, feed = bendecode_int(bits)
        return match, feed
    elif chr(bits[0]).isdigit():
        match, feed = bendecode_str(bits)
        return match, feed
    elif bits.startswith(b"l"):
        lst, feed = bendecode_list(bits)
        return lst, feed
    elif bits.startswith(b"d"):
        dic, feed = bendecode_dict(bits)
        return dic, feed
    else:
        raise DecodeError(bits)


def benencode(val):
    """
    benencode

    Encode data with bencoding.

    Args
    --------
    val : any
        Data for encoding.


    Raises
    --------
    Exception:
        BenencodingError.

    Returns
    ---------
    bytes:
         Bencoded data.
    """
    if type(val) == str:
        return bencode_str(val)
    elif type(val) == int:
        return bencode_int(val)
    elif type(val) == list:
        return bencode_list(val)
    elif type(val) == dict:
        return bencode_dict(val)
    elif hasattr(val, "hex"):
        return bencode_bits(val)
    else:
        raise EncodeError(val)


def bencode_bits(bits):
    """
    bencode_bits

    Encode bytes.

    Args
    --------
    bits : bytes or bytearray
        string

    Returns
    ---------
    bytes:
         bencoded string
    """
    size = str(len(bits)) + ":"
    return size.encode("utf-8") + bits


def bencode_str(txt):
    """
    bencode_str

    Encode string types.

    Args
    --------
    txt (str): string


    Returns
    ---------
    bytes:
         bencoded string
    """
    size = str(len(txt)) + ":"
    return size.encode("utf-8") + txt.encode("utf-8")


def bencode_int(i):
    """
    bencode_int

    Encode integer type.

    Args
    --------
    i : int
        number that needs encoding

    Returns
    ---------
    bytes:
         bencoded int
    """
    return ("i" + str(i) + "e").encode("utf-8")


def bencode_list(elems):
    """
    bencode_list

    Encode list and contents.

    Args
    --------
    elems : list
        list of items for bencoding

    Returns
    ---------
    bytes:
         bencoded list and contents
    """
    arr = bytearray("l", encoding="utf-8")
    for elem in elems:
        encoded = benencode(elem)
        arr.extend(encoded)
    arr.extend(b"e")
    return arr


def bencode_dict(dic):
    """
    bencode_dict

    Encode dictionary and contents.

    Args
    --------
    dic : dict
        key,value pairs of items for bencoding.

    Returns
    ---------
    bytes:
         Bencoded key, value pairs of data.
    """
    result = b"d"
    for k, v in dic.items():
        result += b"".join([benencode(k), benencode(v)])
    return result + b"e"
